#include <linux/limits.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>
#include <errno.h>

#ifndef PORT
#define PORT 10000
#endif

#define MAXLINE 1024
#define ARGSZ 16
#define MAXOBJ 32
#define MAXNUM 100

struct Obj {
	char arg[ARGSZ];
};

struct objentry {
	struct Obj *obj;
	void (*run)(struct Obj *obj);
	int num;
};

struct objentry objects[MAXOBJ];

int findfree();
void EchoObj_run(struct Obj *obj);
void SecretObj_run(struct Obj *obj);
void do_newobj(const char *buf)
{
	int pos, num;
	char type;
	struct Obj *obj;

	pos = findfree();
	if (pos >= MAXOBJ) {
		fprintf(stderr, "full\n");
		return;
	}
	type = *buf++;
	if (type == '\0') {
		fprintf(stderr, "missing object type\n");
		return;
	}
	num = 1;
	if (*buf != '\0') {
		num = atoi(buf);
		if (num <= 0 || num > MAXNUM) {
			fprintf(stderr, "invalid number of objects\n");
			return;
		}
		objects[pos].num = num;
	}
	if ( (obj = calloc(sizeof(*obj), num)) == NULL) {
		fprintf(stderr, "out of memory\n");
		return;
	}
	switch (type) {
	case 'e':
		objects[pos].run = EchoObj_run;
		break;
	case 's':
		objects[pos].run = SecretObj_run;
		break;
	default:
		fprintf(stderr, "invalid type\n");
		free(obj);
		return;
	}
	objects[pos].obj = obj;
	printf("%d\n", pos);
}

// NO INTENTIONAL BUGS AFTER THIS POINT

int Obj_setarg(struct Obj *obj, const char *a)
{
	if (strlen(a) >= ARGSZ)
		return 0;
	strcpy(obj->arg, a);
	return 1;
}

void EchoObj_run(struct Obj *obj)
{
	printf("%s\n", obj->arg);
}

void SecretObj_run(struct Obj *obj)
{
	FILE *f;

	f = fopen("secret", "r");
	if (f == NULL) {
		fprintf(stderr, "error opening secret file\n");
		return;
	}
	if (fread(obj->arg, ARGSZ, 1, f) != 1) {
		fprintf(stderr, "error reading secret file\n");
	}
	fclose(f);
}

int findfree()
{
	struct objentry *e = NULL;
	int i = 0;
	for ( ; i < MAXOBJ; i++)
		if (!objects[i].obj)
			break;
	return i;
}

// expects buf to contain pos[,idx]. If valid,
// sets *ppos and *pidx to pos and idx, respectively.
// Returns a pointer to the first unparsed char, or
// NULL in case of error.
const char *parseref(const char *buf, int *ppos, int *pidx)
{
	char *rest;
	long pos, idx;
	struct Obj *obj;

	if (*buf == '\0') {
		fprintf(stderr, "missing reference\n");
		return NULL;
	}
	errno = 0;
	pos = strtol(buf, &rest, 10);
	if (errno || pos < 0 || pos >= MAXOBJ) {
		fprintf(stderr, "reference out of range\n");
		return NULL;
	}
	if (rest == buf) {
		fprintf(stderr, "invalid input\n");
		return NULL;
	}
	buf = rest;
	obj = objects[pos].obj;
	if (obj == NULL) {
		fprintf(stderr, "invalid reference\n");
		return NULL;
	}
	*ppos = pos;
	idx = 0;
	if (*buf == ',') {
		buf++; // skip the comma
		if (*buf == '\0') {
			fprintf(stderr, "missing index\n");
			return NULL;
		}
		errno = 0;
		idx = strtol(buf, &rest, 10);
		if (errno || idx < 0 || idx >= objects[pos].num) {
			fprintf(stderr, "array index out of range\n");
			return NULL;
		}
		if (rest == buf) {
			fprintf(stderr, "invalid input\n");
			return NULL;
		}
		buf = rest;
	}
	*pidx = idx;
	return buf;
}

void do_setarg(const char *buf)
{
	int pos, idx;
	struct Obj *obj;

	buf = parseref(buf, &pos, &idx);
	if (buf == NULL)
		return;
	if (*buf != '=') {
		fprintf(stderr, "missing arg\n");
		return;
	}
	buf++;
	obj = &objects[pos].obj[idx];
	if (!Obj_setarg(obj, buf)) {
		fprintf(stderr, "not set\n");
	}
}

void do_run(const char *buf)
{
	int pos, idx;
	struct Obj *obj;

	buf = parseref(buf, &pos, &idx);
	if (buf == NULL)
		return;
	if (*buf != '\0') {
		fprintf(stderr, "invalid input\n");
		return;
	}
	obj = &objects[pos].obj[idx];
	objects[pos].run(obj);
}

void child(int uid)
{
	char buf[MAXLINE], *n;

	printf("Secure remote object server v0.9\n");

	for (;;) {
		if (fgets(buf, MAXLINE, stdin) == NULL)
			break;
		n = index(buf, '\n');
		if (n != NULL)
			*n = '\0';
		switch (buf[0]) {
		case '\0':
			break;
		case 'o':
			do_newobj(buf + 1);
			break;
		case 'a':
			do_setarg(buf + 1);
			break;
		case 'r':
			do_run(buf + 1);
			break;
		default:
			fprintf(stderr, "invalid command\n");
			break;
		}
	}
}

void sigchld(int signo)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);
	if (pid > 0) {
		printf("child %ld: ", (long)pid);
		if (WIFEXITED(status)) {
			printf("exited with status: %d\n", WEXITSTATUS(status));
		} else {
			printf("%s\n", strsignal(WTERMSIG(status)));
		}
	}
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, sigchld);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			if (dup(con) < 0)
				goto error;
			close(1);
			if (dup(con) < 0)
				goto error;
			close(2);
			if (dup(con) < 0)
				goto error;
			child(1);
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;

error:
	fprintf(stderr, "dup() failed\n");
	exit(1);
}
